import os

from dotenv import dotenv_values, load_dotenv

from gpyt.free_assistant import FreeAssistant
from gpyt.palm_assistant import PalmAssistant

from .app import gpyt
from .args import USE_EXPERIMENTAL_FREE_MODEL
from .assistant import Assistant
from .config import MODEL, PROMPT

# check for environment variable first
API_KEY = os.getenv("OPENAI_API_KEY")
PALM_API_KEY = os.getenv("PALM_API_KEY")
DOTENV_PATH = os.path.expanduser("~/.env")

if API_KEY is None or (isinstance(API_KEY, str) and not len(API_KEY)):
    result = dotenv_values(DOTENV_PATH)
    API_KEY = result.get("OPENAI_API_KEY", None)
    PALM_API_KEY = result.get("PALM_API_KEY", None)


assert (
    API_KEY is not None and len(API_KEY)
) or USE_EXPERIMENTAL_FREE_MODEL, """

❗Missing OpenAI API Key ❗

Steps to fix this issue:

    1) Get an OpenAI Key from https://platform.openai.com/account/api-keys
    2) Create a `.env` file located in your $HOME directory.
    3) add `OPENAI_API_KEY="your_api_key"` to the `.env` file.
    4) or, you can run `$ EXPORT OPENAI_API_KEY="your_api_key"` to set a key
    for the active shell session.
    5) rerun this program with `$ gpyt` or `$ python -m gpyt`

💡 OR 💡

Use the new **experimental** free model feature.

$ gpyt --free

💭 This may break sometimes and is often times slower. It is recommended to
generate your own OpenAI API key (see above)

"""


gpt = Assistant(api_key=API_KEY or "", model=MODEL, prompt=PROMPT)
gpt4 = Assistant(api_key=API_KEY or "", model="gpt-4", prompt=PROMPT)
free_gpt = FreeAssistant()
palm = PalmAssistant(api_key=PALM_API_KEY)
app = gpyt(assistant=gpt, free_assistant=free_gpt, palm=palm, gpt4=gpt4)



from textual.widgets import ListView
from .conversation_option import StartNewConversationOption, ConversationOption


class VimLikeListView(ListView):
    BINDINGS = [("j", "cursor_down", "Cursor Down"), ("k", "cursor_up", "Cursor Up")]

    def action_select_cursor(self) -> None:
        selected = self.highlighted_child

        assert isinstance(selected, StartNewConversationOption) or isinstance(
            selected, ConversationOption
        ), f"Invalid selection type {type(selected)}"

        selected.select()

    def on_focus(self, _) -> None:
        if self.parent and self.parent.has_class("hidden"):
            user_input = self.app.query_one("#user-input")
            self.app.set_focus(user_input)



import json
import os
from pathlib import Path

from textual import work
from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, LoadingIndicator


from gpyt.free_assistant import FreeAssistant
from gpyt.palm_assistant import PalmAssistant
from ..args import USE_EXPERIMENTAL_FREE_MODEL, USE_PALM_MODEL, USE_GPT4
from ..assistant import Assistant
from ..conversation import Conversation, Message
from ..id import get_id
from .assistant_responses import AssistantResponses
from .options import Options
from .past_conversations import PastConversations
from .user_input import UserInput


class AssistantApp(App):

    """Base app for all user->assistant interactions"""

    BINDINGS = [
        ("ctrl+b", "toggle_dark", "Toggle Dark Mode"),
        ("ctrl+n", "toggle_sidebar", "Past Conversations"),
        ("ctrl+c", "handle_exit", "Quit"),
        ("up", "scroll_convo_up", "Scroll Up Convo"),
        ("down", "scroll_convo_down", "Scroll Down Convo"),
        ("ctrl+o", "toggle_settings", "Settings"),
    ]

    CSS_PATH = "styles.cssx"

    def __init__(
        self,
        assistant: Assistant,
        free_assistant: FreeAssistant,
        palm: PalmAssistant,
        gpt4: Assistant,
    ):
        super().__init__()
        self.assistant = assistant
        self._free_assistant = free_assistant
        self._palm = palm
        self._gpt4 = gpt4
        self.conversations: list[Conversation] = []
        self.active_conversation: Conversation | None = None
        self._convo_ids_added: set[str] = set()
        self.use_free_gpt = USE_EXPERIMENTAL_FREE_MODEL
        self.use_palm = USE_PALM_MODEL
        self.use_gpt4 = USE_GPT4
        self.use_default_model = not (
            self.use_free_gpt or self.use_palm or self.use_gpt4
        )

    def _get_assistant(self) -> Assistant | FreeAssistant | PalmAssistant:
        if self.use_palm:
            return self._palm
        if self.use_free_gpt:
            return self._free_assistant
        if self.use_gpt4:
            return self._gpt4

        return self.assistant

    def adjust_model_border_title(self) -> None:
        model = "GPT 3.5"

        if self.use_free_gpt:
            model = "GPT3.5 Free 🆓"
        elif self.use_palm:
            model = "PaLM 2 🌴"
        elif self.use_gpt4:
            model = "GPT 4"

        self.user_input.border_title = f"Model: {model}"

    def compose(self) -> ComposeResult:
        header = Header(show_clock=True)
        header.tall = False
        yield header
        yield Footer()
        self.assistant_responses = AssistantResponses(app=self)
        self.assistant_responses.border_title = "Conversation History"

        self.user_input = UserInput(app=self)
        self.user_input.border_subtitle = "Press Enter To Submit"
        yield self.user_input
        yield self.assistant_responses
        self.past_conversations = PastConversations(classes="hidden", app=self)
        self.past_conversations.border_title = "Past Conversations"
        self.past_conversations.border_subtitle = "Press Enter to Select"
        self.adjust_model_border_title()
        yield self.past_conversations
        yield Options(classes="hidden", app=self)

    def get_saved_conversations_path(self) -> Path:
        """Return the path where conversations are to be saved/loaded from"""
        GPT_CACHE_DIR = os.getenv("GPT_CACHE_DIR")
        HOME_DIR = os.getenv("HOME")

        if not GPT_CACHE_DIR:
            # if no GPT_CACHE_DIR env provided, default to $HOME... ensure it exists!
            assert HOME_DIR, "Missing $HOME env var"

        return Path(
            GPT_CACHE_DIR if GPT_CACHE_DIR else HOME_DIR,  # type: ignore
            ".cache",
            "gpyt",
            "conversations",
        )

    def load_saved_conversations(self) -> None:
        """Load conversations from saved conversation path"""
        conversations_path = self.get_saved_conversations_path()
        conversation_file_paths = conversations_path.glob(pattern=r"*.json")
        conversation_file_paths = sorted(
            conversation_file_paths, key=lambda f: f.stat().st_mtime
        )

        for path in conversation_file_paths:
            with open(path, "r") as fd:
                raw_json = json.load(fd)
                conversation = Conversation.parse_obj(raw_json)
                self.conversations.append(conversation)
                self._convo_ids_added.add(conversation.id)
                self.past_conversations.add_conversation_option(conversation)

    def save_active_conversation_to_disk(self) -> str:
        """Save the active conversation to disk and return the file path"""
        return self.save_conversation_to_disk()

    def save_conversation_to_disk(
        self, conversation: Conversation | None = None
    ) -> str:
        """
        Save conversation to disk and return the file path.
        Will use active conversation if no conversation is given.
        """
        if not conversation:
            assert (
                self.active_conversation
            ), "When not supplied a conversation to save to disk, there must be an active conversation!"
            conversation = self.active_conversation

        conversations_path = self.get_saved_conversations_path()

        os.makedirs(conversations_path, exist_ok=True)

        path = Path(conversations_path, f"convo-{conversation.id}.json")
        with open(path, "w") as fd:
            json.dump(conversation.dict(), fd)

        self._add_active_as_option()

        return str(path)

    def _set_summary_title_id(self, summary: str, id: str) -> None:
        self.assistant_responses.border_title = f"Conversation History - {summary}"
        self.assistant_responses.border_subtitle = f"convo-id: 0x{id}"

    def _setup_fresh_convo(self, initial_user_input: str) -> None:
        summary = self._get_assistant().get_conversation_summary(initial_user_input)
        new_convo = Conversation(id=get_id(), summary=summary, log=[])
        self._set_summary_title_id(summary, new_convo.id)
        self.conversations.append(new_convo)
        self.active_conversation = new_convo

    def clear_active_conversation(self) -> None:
        """
        Clears the active conversation by removing all messages from the
        current conversation history and resetting it.
        """
        self.assistant_responses.remove()
        self.assistant_responses = AssistantResponses(app=self)
        self.mount(self.assistant_responses)
        self.assistant_responses.border_title = "Conversation History"

    def _add_active_as_option(self) -> None:
        """
        Adds the active conversation as an option to the list of past
        conversations, if it does not already exist.
        """
        if not self.active_conversation:
            return
        exists_already = Path(
            self.get_saved_conversations_path(),
            f"convo-{self.active_conversation.id}.json",
        ).exists()

        if exists_already and self.active_conversation.id not in self._convo_ids_added:
            self.past_conversations.add_conversation_option(self.active_conversation)

        self._convo_ids_added.add(self.active_conversation.id)

    def start_new_conversation(self, add_option: bool = True) -> None:
        """Save current conversation, clear it, start a new fresh one"""
        if not self.active_conversation:
            return

        self.save_active_conversation_to_disk()
        if add_option:
            self._add_active_as_option()
        self.clear_active_conversation()
        self.active_conversation = None
        self.action_toggle_sidebar()

    @work()
    def fetch_assistant_response(self, user_input: str) -> None:
        if self.active_conversation is None:
            self._setup_fresh_convo(user_input)

        assert self.active_conversation, "No active conversation during log write"
        user_message = Message(id=get_id(), role="user", content=user_input)
        self.active_conversation.log.append(user_message)

        try:
            assistant_response_stream = self._get_assistant().get_response_stream(
                user_input
            )
        except:
            assistant_response_stream = self._get_assistant().error_fallback_message

        self.app.call_from_thread(self.assistant_responses.mount, LoadingIndicator())

        self.app.call_from_thread(
            self.assistant_responses.add_response,
            message=user_message,
            stream=assistant_response_stream,
        )

    def on_select_previous_conversation(self, conversation: Conversation) -> None:
        self.start_new_conversation(add_option=False)
        self.assistant_responses.setup_from_presaved_conversation(conversation)
        self.assistant_responses.container.scroll_end(duration=2.0, easing="in_quart")

    def action_toggle_dark(self) -> None:
        self.dark = not self.dark

    def action_toggle_sidebar(self) -> None:
        if not self.past_conversations.has_class("opened-gt-once"):
            try:
                self.load_saved_conversations()
            except:
                self.past_conversations.add_conversation_option(
                    Conversation(
                        summary=f"Failed to load conversations from {self.get_saved_conversations_path()}",
                        id="-1",
                        log=[],
                    )
                )
        self.past_conversations.add_class("opened-gt-once")
        self.past_conversations.toggle_class("hidden")
        if self.past_conversations.has_class("hidden"):
            self.focus_user_input()
            return
        self.set_focus(self.past_conversations.options)

    def action_toggle_settings(self) -> None:
        options = self.query_one(Options)
        options.toggle_class("hidden")
        if options.has_class("hidden"):
            self.focus_user_input()
            return
        self.set_focus(options)

    def focus_user_input(self) -> None:
        inp = self.query_one("#user-input")
        inp.focus()

    def action_handle_exit(self) -> None:
        exit()

    def action_scroll_convo_up(self) -> None:
        self.assistant_responses.container.scroll_relative(y=-4)

    def action_scroll_convo_down(self) -> None:
        self.assistant_responses.container.scroll_relative(y=4)



SUMMARY_PROMPT = """Summarize the following question in 6 words or less. be
concise and fully grasp the main idea of the question. If what follows
is not a question just summarize the idea of the statement itself\n"""


API_ERROR_MESSAGE = """
❗ There was an error with OpenAI.

# Diagnostics

Make sure that:
1) You have a proper OPENAI_API_KEY configured at `$HOME/.env` starting with 'sk-'

2) You have configured Billing over at https://platform.openai.com/account/billing

3) You aren't getting rate limited (9000 tokens per minute)

4) If trying to use GPT4, be sure you have been granted access from OpenAI

Try Again Later.
"""

API_ERROR_FALLBACK = [{"choices": [{"delta": {"content": API_ERROR_MESSAGE}}]}]


INTRO = "Ask me anything. I'll try to assist you!"

AVAILABLE_MODELS = ["gpt-3.5-turbo"]

MODEL = "gpt-3.5-turbo"

PROMPT = """
You are a helpful assistant. Be accurate and concise with your responses.
You are being used as a command-line tool by the user and you may be asked
technical/programming type questions.

Use basic Markdown syntax.

Use # for big ideas or ## for smaller ideas.

Use lists with "*" or "1), 2), ... n)" when it makes sense to list things (like stages or parts)

For intermediate to complex topics (like science, or large processes) only you should
summarize the user's question into less than four words, and
place them in a markdown header as the first line of your response.

For example, if the user is asking about football use this format

# Football

<the rest of your response>

--end intermediate to complex topics instructions--
"""



from textual import work
from textual.app import ComposeResult
from textual.containers import ScrollableContainer
from textual.widgets import LoadingIndicator, Static
from typing import Generator

from ..conversation import Conversation, Message
from ..id import get_id
from .assistant_response import AssistantResponse


class AssistantResponses(Static):
    """Container for individual AssistantResponse widgets"""

    def __init__(self, app):
        super().__init__()
        self.container = ScrollableContainer()
        self._app = app

    def compose(self) -> ComposeResult:
        yield self.container

    def setup_from_presaved_conversation(self, conversation: Conversation) -> None:
        """Load conversation into conversation history widget from Conversation model"""

        self._app.active_conversation = conversation
        self._app._set_summary_title_id(conversation.summary, conversation.id)
        new_history = []

        all_user_messages = [m for m in conversation.log if m.role == "user"]
        all_assistant_messages = [m for m in conversation.log if m.role == "assistant"]

        for user_message, assistant_response in zip(
            all_user_messages, all_assistant_messages
        ):
            assert (
                user_message.role == "user"
            ), "Improper role for setup from presaved convesation"

            new_response = AssistantResponse(
                question=user_message.content, id=user_message.id
            )
            self.container.mount(new_response)
            assert (
                assistant_response.role == "assistant"
            ), "Improper role for setup from presaved convesation"
            new_response.update_response(assistant_response.content)
            new_history.append({"role": "user", "content": user_message.content})
            new_history.append(
                {"role": "assistant", "content": assistant_response.content}
            )

        self._app._get_assistant().set_history(new_history)
        self._app.past_conversations.add_class("hidden")
        self._app.focus_user_input()

    @work()
    def add_response(self, stream: Generator, message: Message) -> None:
        new_response = AssistantResponse(question=message.content, id=message.id)
        self._app.call_from_thread(self.container.mount, new_response)
        self._app.call_from_thread(new_response.scroll_visible)
        markdown = ""
        update_frequency = 10
        i = 0
        for data in stream:
            i += 1
            try:  # HACK: should check for attr, not try/except
                if self._app.use_free_gpt or self._app.use_palm:
                    markdown = markdown + data
                else:
                    markdown = markdown + data["choices"][0]["delta"]["content"]
                if i % update_frequency == 0:
                    self._app.call_from_thread(new_response.update_response, markdown)
                    self._app.call_from_thread(self.container.scroll_end)
            except:
                continue

        self._app.call_from_thread(new_response.update_response, markdown)
        self._app.call_from_thread(
            new_response.user_question.scroll_visible, duration=2, easing="out_back"
        )
        self._app._get_assistant().log_assistant_response(markdown)
        assistant_message = Message(id=get_id(), role="assistant", content=markdown)
        assert self._app.active_conversation, "No active conversation during log write"
        self._app.active_conversation.log.append(assistant_message)

        self._app.call_from_thread(self._app.save_active_conversation_to_disk)

        loading_indicator = self.query_one(LoadingIndicator)
        if loading_indicator:
            self._app.call_from_thread(loading_indicator.remove)
            print("deleting loading indicator")



from textual.widgets import Label, ListItem
from textual.app import ComposeResult
from ..conversation import Conversation


class StartNewConversationOption(ListItem):
    def __init__(self, app):
        super().__init__()
        self._app = app

    def compose(self) -> ComposeResult:
        yield Label("<-- Start a new conversation -->", id="new-convo-option")

    def watch_highlighted(self, value: bool) -> None:
        super().watch_highlighted(value)
        if value:
            self.add_class("highlighted")
        else:
            self.remove_class("highlighted")

    def select(self) -> None:
        self._app.start_new_conversation()


class ConversationOption(ListItem):
    ELLIPSIFY_CUTOFF = 35

    def __init__(self, conversation: Conversation, app):
        super().__init__()
        self.conversation = conversation
        self._app = app

    def compose(self) -> ComposeResult:
        yield Label(self._ellipsify_long_summary(self.conversation.summary))

    def _ellipsify_long_summary(self, summary: str) -> str:
        if len(summary) < ConversationOption.ELLIPSIFY_CUTOFF:
            return summary

        summary = summary[: ConversationOption.ELLIPSIFY_CUTOFF - 2] + "..."
        return summary

    def select(self) -> None:
        self._app.on_select_previous_conversation(self.conversation)
        # app.assistant_responses.setup_from_presaved_conversation(self.conversation)



import os
from typing import Generator

import google.generativeai as palm

from .assistant import Assistant
from .config import API_ERROR_FALLBACK, PROMPT


class PalmAssistant(Assistant):
    EXAMPLES = [
        ("How far is the sun?", "Distance to Sun"),
        ("How do I make a cake if I have some ingredients", "Baking a Cake"),
        ("Write a for loop in C#", "C# For Loop"),
        ("Hello!", "Greeting"),
    ]

    SUMMARY_PROMPT = """Summarize the following question in 6 words or less. be
        concise and fully grasp the main idea of the question. If what follows
        is not a question just summarize the idea of the statement itself


        input: How far is the moon?
        summary Distance to Moon?
        input: Write a for loop in C#
        summary C# For Loops.
        input: How do I bake a cake?
        summary Baking a Cake.
    """

    API_ERROR_MESSAGE = """There was an ERROR with the PaLM 2 API\n## Diagnostics\n* Make sure you have a PALM_API_KEY set at `~/.env`\n* Make sure you aren't being rate-limited."""

    def __init__(self, api_key):
        self.error_fallback_message = API_ERROR_FALLBACK
        self.chat: list[dict[str, str]] = []
        self.api_key = api_key
        if not self._bad_key():
            palm.configure(api_key=self.api_key)
        self.messages = []
        self.error_fallback_message = PalmAssistant.API_ERROR_MESSAGE

    def set_history(self, new_history: list[dict[str, str]]):
        self.clear_history()
        for message in new_history:
            self.messages.append(message["content"])

    def clear_history(self) -> None:
        """reset internal message queue"""
        self.messages.clear()

    def get_response_stream(self, user_input: str) -> Generator:
        """Fake a stream output with PaLM 2"""
        response = self.get_response(user_input)
        for i in range(0, len(response), 8):
            yield response[i : i + 8]

    def log_assistant_response(self, final_response: str) -> None:  # pyright: ignore
        ...

    def get_response(self, user_input: str) -> str:
        if self._bad_key():
            return PalmAssistant.API_ERROR_MESSAGE
        self.messages.append(user_input)
        response = palm.chat(context=PROMPT, messages=self.messages).last
        self.messages.append(response)
        return response

    def _bad_key(self) -> bool:
        return not self.api_key or (type(self.api_key) is str and len(self.api_key) < 1)

    def get_conversation_summary(self, initial_message: str) -> str:
        if self._bad_key():
            return "API KEY Error"
        prompt = f"{PalmAssistant.SUMMARY_PROMPT}\ninput: {initial_message}\nsummary"
        response = palm.generate_text(prompt=prompt)
        return response.result


if __name__ == "__main__":
    ...
    # agent = PalmAssistant()
    # print(agent.get_conversation_summary("What is the third closest moon from Saturn"))
    # while 1:
    #     user_inp = input("> ")
    #     msg = agent.get_response(user_input=user_inp)
    #     print("\n", msg)



from typing import Callable

from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Input, Label

from ..config import INTRO


class UserInput(Container):
    """
    User-facing input box

    Handles special keywords by mapping to callbacks handling special behavior

    save -> save the current conversation to disk
    """

    def __init__(self, app):
        super().__init__()
        self._app = app
        self.keyword_mappings: dict[str, Callable] = {
            # "save": app.save_active_conversation_to_disk,
            "new": self._app.start_new_conversation,
            "clear": self._app.start_new_conversation,
        }

    def compose(self) -> ComposeResult:
        yield Label(f"🤖: {INTRO}", id="help-text")
        self.inp = Input(placeholder="How far away is the Sun?", id="user-input")
        self.inp.focus()
        yield self.inp

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        user_input = event.input.value
        event.input.value = ""

        # empty queries are useless
        if len(user_input) < 1:
            return

        keyword_callback = self.keyword_mappings.get(user_input.strip().lower(), None)
        if keyword_callback:
            keyword_callback()
            return  # don't give anything to the assistant if we use a special callback

        self._app.fetch_assistant_response(user_input)



from textual.app import ComposeResult
from textual.widgets import RadioButton, RadioSet


class Options(RadioSet):
    BINDINGS = [
        ("j", "next_button", "Cursor Down"),
        ("k", "previous_button", "Cursor Up"),
    ]

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

    def compose(self) -> ComposeResult:
        self.border_title = "Options"
        self.border_subtitle = "Use Space to Select"

        self.use_default = RadioButton(
            "Use GPT 3.5 (paid, default)", id="use_default_model"
        )
        self.use_free = RadioButton("Use Free Model (experimental)", id="use_free_gpt")
        self.use_palm = RadioButton("Use PaLM 2 (Google)", id="use_palm")
        self.use_gpt4 = RadioButton("Use GPT4 (Access Required)", id="use_gpt4")

        yield self.use_default
        yield self.use_free
        yield self.use_palm
        yield self.use_gpt4

    def on_radio_set_changed(self, event):
        option = event.pressed.id
        if "use" in option:
            self._app.use_free_gpt = False
            self._app.use_default_model = False
            self._app.use_palm = False
            self._app.use_gpt4 = False

        if hasattr(self._app, option):
            self._app.__dict__[option] = True
            self._app.adjust_model_border_title()



import pyperclip
from textual import work
from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label, Markdown, Static


class AssistantResponse(Static):
    """Each User/Assistant interaction"""

    def __init__(self, question: str = "", id: str = ""):
        super().__init__()
        self.question = question
        self._id = id
        self.response_view = Markdown()
        self._last_content = ""

    def compose(self) -> ComposeResult:
        self.user_question = Label(f"😀: {self.question}", classes="convo")
        yield self.user_question
        yield Label("🤖:", classes="convo")
        container = Container(self.response_view, id="response-container")
        container.border_subtitle = f"message-id: 0x{self._id}"
        yield container

    def on_click(self) -> None:
        pyperclip.copy(self._last_content)

    @work()
    def update_response(self, content: str) -> None:
        self.app.call_from_thread(self.response_view.update, content)
        self._last_content = content



from textual.containers import Container
from textual.app import ComposeResult
from .conversation_option import StartNewConversationOption, ConversationOption
from .vim_list import VimLikeListView
from ..conversation import Conversation


class PastConversations(Container):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

    def compose(self) -> ComposeResult:
        self.start_new_conversation_option = StartNewConversationOption(app=self._app)
        self.options = VimLikeListView(self.start_new_conversation_option)
        yield self.options

    def add_conversation_option(self, conversation: Conversation) -> None:
        option = ConversationOption(conversation, app=self._app)
        self.options.mount(option, before=1)



import argparse

parser = argparse.ArgumentParser()
parser.add_argument(
    "--free", help="Use the gpt4free model. (experimental)", action="store_true"
)
parser.add_argument(
    "--palm", help="Use PaLM 2, Google's Premier LLM (new)", action="store_true"
)

parser.add_argument(
    "--gpt4",
    help="Use GPT4 (your API_KEY must have been granted access.)",
    action="store_true",
)


args = parser.parse_args()

USE_EXPERIMENTAL_FREE_MODEL = args.free
USE_PALM_MODEL = args.palm
USE_GPT4 = args.gpt4
